package edu.tum.motras.algorithms;

import static edu.tum.motras.algorithms.util.Util.checkArguments;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import com.google.common.collect.ImmutableSet;

import edu.tum.motras.algorithms.util.StatePair;
import edu.tum.motras.algorithms.util.Util;
import edu.tum.xmts.Action;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;

public class Conjunction extends WaitingQueueAlgorithm<StatePair> implements Algorithm<State> {

	private Set<StatePair> relation;

	@Override
	public State apply(State... input) {
		checkArguments(this, input);

		ConsistencyRelation consistencyRelation = new ConsistencyRelation();

		relation = consistencyRelation.apply(input[0], input[1]);
		StatePair firstPair = new StatePair(input[0], input[1]);

		if (!relation.contains(firstPair)) {
			return null;
		}

		run(firstPair, relation);

		relation = null;

		return transitionSystem.states.getElements().get(0);
	}

	@Override
	public int getArity() {
		return 2;
	}

	@Override
	public Set<Type> getSupportedTypes() {
		return ImmutableSet.of(Type.MTS, Type.DMTS);
	}

	@Override
	public String toString() {
		return "Conjunction";
	}

	private State getState(State state1, State state2) {
		StatePair pair = new StatePair(state1, state2);

		if (!relation.contains(pair)) {
			return null;
		}

		return stateMapper.map(pair);
	}

	@Override
	protected void process(StatePair statePair) {
		State newState = stateMapper.map(statePair);
		State state1 = statePair.get(1);
		State state2 = statePair.get(2);

		/* Construct Transitions */
		for (Transition transition1 : state1.getTransitions()) {
			Action action = transition1.getAction();

			for (Transition transition2 : state2.getTransitions(action.getId())) {
				State nextState = getState(transition1.getState(), transition2.getState());

				if (nextState != null) {
					newState.createTransition(actionMapper.map(action), nextState);
				}
			}
		}

		Set<Collection<Transition>> disjMust = new HashSet<>();

		/* Construct Obligation */
		for (Set<Transition> mustTransitions : Util.getDisjunctiveMustTransitions(state1)) {
			for (Transition transition1 : mustTransitions) {
				Action action = transition1.getAction();

				Collection<Transition> mayTransitions = state2.getTransitions(action.getId());

				if (mayTransitions.isEmpty()) {
					continue;
				}

				Set<Transition> dTransitions = new HashSet<>();

				for (Transition transition2 : mayTransitions) {
					State nextState = getState(transition1.getState(), transition2.getState());

					if (nextState != null) {
						dTransitions.add(newState.getTransition(actionMapper.map(action), nextState));
					}
				}

				disjMust.add(dTransitions);
			}
		}

		for (Set<Transition> mustTransitions : Util.getDisjunctiveMustTransitions(state2)) {
			for (Transition transition2 : mustTransitions) {
				Action action = transition2.getAction();

				Collection<Transition> mayTransitions = state1.getTransitions(action.getId());

				if (mayTransitions.isEmpty()) {
					continue;
				}

				Set<Transition> dTransitions = new HashSet<>();

				for (Transition transition1 : mayTransitions) {
					State nextState = getState(transition1.getState(), transition2.getState());

					if (nextState != null) {
						dTransitions.add(newState.getTransition(actionMapper.map(action), nextState));
					}
				}

				disjMust.add(dTransitions);
			}
		}

		newState.setDisjunctiveMustObligation(disjMust);
	}

	@Override
	protected void setup(TransitionSystem system) {
		system.setId("conjunction");
		system.setType(Type.DMTS);
	}

	@Override
	public boolean isConfigurable() {
		return false;
	}

	@Override
	public void setConfiguration(String conf) {
		throw new UnsupportedOperationException("This algorithm cannot be configured");
	}
}
