package edu.tum.motras.algorithms;

import static edu.tum.motras.algorithms.util.Util.checkArguments;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeoutException;
import java.util.logging.Logger;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Table;

import edu.tum.motras.algorithms.util.StatePair;
import edu.tum.motras.algorithms.util.Util;
import edu.tum.motras.legacy.pmtsrc.util.QDIMACS;
import edu.tum.motras.legacy.pmtsrc.util.QDIMACS.QSatSolverType;
import edu.tum.motras.legacy.pmtsrc.util.formula.AbstractLiteralFactory;
import edu.tum.motras.legacy.pmtsrc.util.formula.AbstractLiteralFactory.Domain;
import edu.tum.motras.legacy.pmtsrc.util.formula.LiteralFactory;
import edu.tum.motras.legacy.pmtsrc.util.formula.PrefixedTransitionFactory;
import edu.tum.motras.legacy.pmtsrc.util.formula.QuantifiedBoolean;
import edu.tum.motras.legacy.pmtsrc.util.formula.TreeBoolean;
import edu.tum.xmts.Parameter;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.formula.Formula;
import edu.tum.xmts.formula.Formula.Atom;
import edu.tum.xmts.formula.Formula.Conjunction;
import edu.tum.xmts.formula.Formula.Disjunction;
import edu.tum.xmts.formula.Formula.Equivalence;
import edu.tum.xmts.formula.Formula.Implication;
import edu.tum.xmts.formula.Formula.Negation;
import edu.tum.xmts.formula.Visitor;

public class ModalRefinementQSat implements Algorithm<Boolean> {

	@Override
	public boolean isConfigurable() {
		return true;
	}

	@Override
	public void setConfiguration(String conf) {
		try {
			QSat = new QDIMACS(QSatSolverType.valueOf(conf));
		} catch (IOException e) {
			Logger.getGlobal().severe("Could not change to the specified solver");
		}
	}

	public QDIMACS getInternalSolver() {
		return QSat;
	}

	private QDIMACS QSat;

	private TransitionSystem abstraction;
	private TransitionSystemLiteralFactory abstractionFactory;

	private TransitionSystem refinement;
	private TransitionSystemLiteralFactory refinementFactory;

	private PrefixedTransitionFactory prefixedTransitionLibrary;

	private Collection<StatePair> modalRefinemenRelation;
	private Table<State, State, StatePair> relation;
	private LiteralFactory<StatePair> relationLibrary;

	public ModalRefinementQSat() throws IOException {
		QSat = new QDIMACS();
	}

	@Override
	public Boolean apply(State... input) throws IllegalArgumentException, TimeoutException, IOException {
		checkArguments(this, input);
		return solve(input[0], input[1]);
	}

	@Override
	public int getArity() {
		return 2;
	}

	@Override
	public Set<Type> getSupportedTypes() {
		return ImmutableSet.of(Type.MTS, Type.DMTS, Type.BMTS, Type.PMTS);
	}

	@Override
	public String toString() {
		return "Modal Refinement (QSat)";
	}

	private Boolean solve(State refiningState, State abstractState) throws IOException, TimeoutException {
		refinement = refiningState.getContext();
		abstraction = abstractState.getContext();

		refinementFactory = new TransitionSystemLiteralFactory(refinement);
		abstractionFactory = new TransitionSystemLiteralFactory(abstraction);

		prefixedTransitionLibrary = abstractionFactory.createPrefixedTransitionLibrary();
		modalRefinemenRelation = Util.getReachables(refiningState, abstractState);

		relation = HashBasedTable.create();
		relationLibrary = new LiteralFactory<>(Domain.RELATION, modalRefinemenRelation);

		for (StatePair statePair : modalRefinemenRelation) {
			relation.put(statePair.get(1), statePair.get(2), statePair);
		}

		prefixedTransitionLibrary.startInitialization();
		TreeBoolean treeFormula = createFormula(relation.get(refiningState, abstractState));
		prefixedTransitionLibrary.finishInitialization();
		QuantifiedBoolean quantifiedFormula;

		quantifiedFormula = new QuantifiedBoolean(treeFormula);

		quantifiedFormula.exists(prefixedTransitionLibrary.getLiteralRange());
		quantifiedFormula.forall(refinementFactory.transitionLibrary.getLiteralRange());
		quantifiedFormula.exists(relationLibrary.getLiteralRange());
		quantifiedFormula.exists(abstractionFactory.parameterLibrary.getLiteralRange());
		quantifiedFormula.forall(refinementFactory.parameterLibrary.getLiteralRange());

		/* Backend Execution */
		Boolean result = QSat.solve(quantifiedFormula);

		AbstractLiteralFactory.deinitializeAll();

		return result;
	}

	private TreeBoolean createFormula(StatePair entry) {
		final ArrayList<TreeBoolean> clauses = new ArrayList<>();

		clauses.add(relationLibrary.getLiteral(entry));

		for (StatePair st : modalRefinemenRelation) {
			clauses.add(createRelationClause(st));
		}

		return TreeBoolean.and(clauses);
	}

	private TreeBoolean createMatchingCondition(StatePair st) {
		final State refiningState = st.get(1);
		final State abstractState = st.get(2);
		final List<TreeBoolean> con = new LinkedList<>();

		for (Transition sa : refiningState.getTransitions()) {
			final String actionId = sa.getAction().getId();
			List<TreeBoolean> dis = new LinkedList<>();

			for (Transition ta : abstractState.getTransitions(actionId)) {
				TreeBoolean tAtom = prefixedTransitionLibrary.getLiteral(st, ta);
				TreeBoolean rAtom = relationLibrary.getLiteral(relation.get(sa.getState(), ta.getState()));
				dis.add(tAtom.and(rAtom));
			}

			TreeBoolean tAtom = refinementFactory.transitionLibrary.getLiteral(sa);

			if (!dis.isEmpty()) {
				con.add(tAtom.implies(TreeBoolean.or(dis)));
			} else {
				con.add(tAtom.not());
			}
		}

		for (Transition ta : abstractState.getTransitions()) {
			final String actionId = ta.getAction().getId();
			List<TreeBoolean> dis = new LinkedList<>();

			for (Transition sa : refiningState.getTransitions(actionId)) {
				TreeBoolean tAtom = refinementFactory.transitionLibrary.getLiteral(sa);
				TreeBoolean rAtom = relationLibrary.getLiteral(relation.get(sa.getState(), ta.getState()));
				dis.add(tAtom.and(rAtom));
			}

			TreeBoolean tAtom = prefixedTransitionLibrary.getLiteral(st, ta);

			if (!dis.isEmpty()) {
				con.add(tAtom.implies(TreeBoolean.or(dis)));
			} else {
				con.add(tAtom.not());
			}
		}

		return TreeBoolean.and(con);
	}

	private TreeBoolean createRelationClause(StatePair st) {
		TreeBoolean relationCondition = createRelationCondition(st);
		TreeBoolean clause = relationLibrary.getLiteral(st).implies(relationCondition);

		return clause;
	}

	private TreeBoolean createRelationCondition(StatePair st) {
		TreeBoolean obligationLeft = refinementFactory.obligations.get(st.get(1)).clone();
		TreeBoolean obligationRight = prefixedTransitionLibrary.prefix(st, abstractionFactory.obligations
				.get(st.get(2)).clone());
		TreeBoolean matchingCondition = createMatchingCondition(st);
		TreeBoolean condition = obligationLeft.implies(obligationRight.and(matchingCondition));

		return condition;
	}
}

class TransitionSystemLiteralFactory {

	protected class CompatVisitior implements Visitor<TreeBoolean> {

		@Override
		public TreeBoolean visit(Atom<?> atom) {
			Object object = atom.getPayload();

			if (atom == Formula.TRUE) {
				return AbstractLiteralFactory.getTrue();
			}

			if (atom == Formula.FALSE) {
				return AbstractLiteralFactory.getFalse();
			}

			if (object instanceof Parameter) {
				Parameter parameter = (Parameter) object;
				return parameterLibrary.getLiteral(parameter);
			}

			if (object instanceof Transition) {
				Transition transition = (Transition) object;
				return transitionLibrary.getLiteral(transition);
			}

			System.err.println("Unknown payload " + (object != null ? object.toString() : "null"));
			System.err.println("Returning true.");
			return AbstractLiteralFactory.getTrue();
		}

		@Override
		public TreeBoolean visit(Conjunction conjunction) {
			return TreeBoolean.and(visitComposite(conjunction.getChildren()));
		}

		@Override
		public TreeBoolean visit(Disjunction disjunction) {
			return TreeBoolean.or(visitComposite(disjunction.getChildren()));
		}

		@Override
		public TreeBoolean visit(Equivalence equivalence) {
			TreeBoolean left = equivalence.getChildren().get(0).accept(this);
			TreeBoolean right = equivalence.getChildren().get(1).accept(this);
			return left.iff(right);
		}

		@Override
		public TreeBoolean visit(Implication implication) {
			TreeBoolean left = implication.getChildren().get(0).accept(this);
			TreeBoolean right = implication.getChildren().get(1).accept(this);
			return left.implies(right);
		}

		@Override
		public TreeBoolean visit(Negation negation) {
			TreeBoolean child = negation.getChildren().get(0).accept(this);
			return child.not();
		}

		private List<TreeBoolean> visitComposite(Collection<Formula> formulas) {
			List<TreeBoolean> treeBooleans = new LinkedList<>();

			for (Formula formula : formulas) {
				treeBooleans.add(formula.accept(this));
			}

			return treeBooleans;
		}
	}

	final Map<State, TreeBoolean> obligations;
	final LiteralFactory<Parameter> parameterLibrary;
	final LiteralFactory<Transition> transitionLibrary;

	public TransitionSystemLiteralFactory(TransitionSystem transitionSystem) {
		parameterLibrary = new LiteralFactory<>(Domain.PARAMETER, transitionSystem.parameters.getElements());
		transitionLibrary = new LiteralFactory<>(Domain.TRANSITION);
		obligations = new HashMap<>();

		CompatVisitior compatVisitior = new CompatVisitior();

		transitionLibrary.startInitialization();

		for (State state : transitionSystem.states.getElements()) {
			for (Transition transition : state.getTransitions()) {
				transitionLibrary.getLiteral(transition);
			}

			obligations.put(state, state.getObligation().getFormula().accept(compatVisitior));
		}

		transitionLibrary.finishInitialization();
	}

	public PrefixedTransitionFactory createPrefixedTransitionLibrary() {
		return new PrefixedTransitionFactory(transitionLibrary);
	}
}