package edu.tum.motras.algorithms;

import static com.google.common.base.Preconditions.checkArgument;
import static edu.tum.motras.algorithms.util.Util.checkArguments;

import java.util.Collection;
import java.util.Deque;
import java.util.EnumSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableList.Builder;
import com.google.common.collect.SetMultimap;

import edu.tum.motras.algorithms.util.StatePair;
import edu.tum.motras.algorithms.util.Util;
import edu.tum.xmts.Action;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;

public class Quotient extends WaitingQueueAlgorithm<StatePair> implements Algorithm<State> {

	private State trapState;
	private List<String> alphabet;
	private Deque<State> inconsitentStates;
	private SetMultimap<State, State> mustParents;

	@Override
	public State apply(State... states) throws IllegalArgumentException {
		checkArguments(this, states);

		/* Build Common Alphabet */
		Builder<String> alphabetBuilder = ImmutableList.builder();

		for (Action action : states[0].getContext().actions.getElements()) {
			alphabetBuilder.add(action.getId());
		}

		for (Action action : states[1].getContext().actions.getElements()) {
			alphabetBuilder.add(action.getId());
		}

		alphabet = alphabetBuilder.build();

		inconsitentStates = new LinkedList<>();
		mustParents = HashMultimap.create();
		run(new StatePair(states[0], states[1]));

		/*
		 * After doing the construction, the pruning takes place, i.e. for each
		 * inconsistent node s, delete s and all the nodes s' that can reach s
		 * using must transitions and of course, remove (may) transitions to
		 * these nodes (s and all s')
		 */

		/* Pruning */
		while (!inconsitentStates.isEmpty()) {
			State inconsitentState = inconsitentStates.remove();
			inconsitentStates.addAll(mustParents.get(inconsitentState));
			transitionSystem.states.remove(inconsitentState);
		}

		if (transitionSystem.getInitialState() != null) {
			Util.removeUnreachableStates(transitionSystem, transitionSystem.getInitialState());
		}

		inconsitentStates = null;
		mustParents = null;
		alphabet = null;
		trapState = null;

		return transitionSystem.getInitialState();
	}

	@Override
	public int getArity() {
		return 2;
	}

	@Override
	public Set<Type> getSupportedTypes() {
		return EnumSet.of(Type.MTS);
	}

	@Override
	public String toString() {
		return "Quotient";
	}

	private Transition getTransition(State context, String action, State state) {
		Transition transition = context.getTransition(transitionSystem.actions.lookup(action), state);

		if (transition == null) {
			transition = context.createTransition(transitionSystem.actions.lookup(action), state);
		}

		return transition;
	}

	@Override
	protected void process(StatePair element) {
		checkArgument(element.get(1).getActions().size() == element.get(1).getTransitions().size(),
				"State %s is non-determinstic.", element.get(1).getId());
		checkArgument(element.get(2).getActions().size() == element.get(2).getTransitions().size(),
				"State %s is non-determinstic.", element.get(2).getId());

		State state = stateMapper.map(element);
		List<StatePair> localQueue = new LinkedList<>();
		List<Transition> mustTransitions = new LinkedList<>();

		for (String action : alphabet) {
			Collection<Transition> transitions1 = element.get(1).getTransitions(action);
			Collection<Transition> transitions2 = element.get(2).getTransitions(action);
			Collection<Transition> mustTransitions1 = Util.getMustTransitions(element.get(1), action);
			Collection<Transition> mustTransitions2 = Util.getMustTransitions(element.get(2), action);

			if (!mustTransitions1.isEmpty() && mustTransitions2.isEmpty()) {
				inconsitentStates.add(state);
				return;
			}

			if (transitions2.isEmpty()) {
				getTransition(state, action, trapState);
			} else if (!transitions1.isEmpty()) {
				Transition transition1 = transitions1.iterator().next();
				Transition transition2 = transitions2.iterator().next();

				StatePair pair = new StatePair(transition1.getState(), transition2.getState());
				localQueue.add(pair);
				State nextState = stateMapper.map(pair);
				Transition transition = getTransition(state, action, nextState);

				if (!mustTransitions1.isEmpty() && !mustTransitions2.isEmpty()) {
					mustTransitions.add(transition);
					mustParents.put(nextState, state);
				}
			}
		}

		for (StatePair pair : localQueue) {
			queue(pair);
		}

		state.setMustObligation(mustTransitions);
	}

	@Override
	protected void setup(TransitionSystem system) {
		system.setId("Quotient");
		trapState = system.states.create("trap");

		for (String actionId : alphabet) {
			if (system.actions.contains(actionId)) {
				continue;
			}

			Action action = system.actions.create(actionId);
			trapState.createTransition(action, trapState);
		}
	}

	@Override
	public boolean isConfigurable() {
		return false;
	}

	@Override
	public void setConfiguration(String conf) {
		throw new UnsupportedOperationException("This algorithm cannot be configured");
	}

}
