package edu.tum.motras.algorithms.cli;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.SequenceInputStream;
import java.util.List;
import java.util.logging.Logger;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;

import edu.tum.motras.algorithms.Algorithm;
import edu.tum.motras.benchmarker.BenchmarkSuiteExecutor;
import edu.tum.xmts.State;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.utils.XMTSReader;
import edu.tum.xmts.utils.XMTSReader.MalformedInputException;
import edu.tum.xmts.utils.XMTSWriter;

public class CLI {

	private static final String classpath = "edu.tum.motras.algorithms.";
	private static final Logger logger = Logger.getLogger(XMTSReader.class.getName());

	private static List<TransitionSystem> transitionSystems;

	@SuppressWarnings("unchecked")
	public static void main(String[] args) {

		/* CLI options */
		Options options = new Options();

		options.addOption("a", "algorithm", true, "Specify the used algorithm");
		options.addOption("p", "parameters", true, "Specify additional parameters for the algorithm");
		options.addOption("q", "query", false, "Display the arity and a list of supported types for the algorithm");
		options.addOption("h", "help", false, "Prints this help message");

		options.addOption("b", "benchmark", true, "Use the given benchmark suite to benchmark the specified algorithm");

		CommandLine cmd = null;

		try {
			cmd = (new PosixParser()).parse(options, args);
		} catch (ParseException e) {
			logger.severe(e.getMessage());
			printUsage(options);
			System.exit(-1);
			return;
		}

		if (args.length == 0 || cmd.hasOption("h") || !cmd.hasOption("a")) {
			printUsage(options);
			System.exit(-1);
		}

		/* Get Algorithm Object */

		try {
			Algorithm<?> algorithm = getAlgorithm(cmd);

			if (cmd.hasOption("q")) {
				System.out.println("Arity: " + algorithm.getArity());
				System.out.println("Supported Types: " + algorithm.getSupportedTypes().toString());
			} else if (cmd.hasOption("b")) {
				BenchmarkSuiteExecutor.runBenchmarkSuite((Algorithm<Boolean>) algorithm,
						new File(cmd.getOptionValue("b")), false);
			} else {
				State input[] = getInput(cmd);
				Object output = algorithm.apply(input);
				setOutput(output);
			}
		} catch (ClassNotFoundException e) {
			System.out.println("The algorithm " + cmd.getOptionValue("a") + " cannot be found");
			System.exit(-1);
		} catch (Exception e) {
			logger.severe(e.getMessage());
			e.printStackTrace();
			System.exit(-1);
		}

		System.exit(0);
	}

	private static Algorithm<?> getAlgorithm(CommandLine cmd) throws Exception {
		String className = classpath + cmd.getOptionValue("a");
		@SuppressWarnings("unchecked")
		Class<Algorithm<?>> algorithmClass = (Class<Algorithm<?>>) Class.forName(className);
		Algorithm<?> algorithm = algorithmClass.newInstance();

		if (cmd.hasOption("p") && algorithm.isConfigurable()) {
			algorithm.setConfiguration(cmd.getOptionValue("p"));
		}

		return algorithm;
	}

	/**
	 * @param cmd
	 */
	private static State[] getInput(CommandLine cmd) throws MalformedInputException, IOException {
		XMTSReader reader = new XMTSReader();

		// Weird as hell, but it works now...
		InputStream antlrWorkAround = new ByteArrayInputStream("mts a [state s]\n".getBytes("UTF-8"));
		SequenceInputStream inputStream = new SequenceInputStream(antlrWorkAround, System.in);

		transitionSystems = reader.readList(inputStream);

		State[] input = new State[transitionSystems.size() - 1];

		// Drop workaround system.
		for (int i = 1; i < transitionSystems.size(); i++) {
			input[i - 1] = transitionSystems.get(i).getInitialState();
		}

		return input;
	}

	private static void printUsage(Options options) {
		HelpFormatter usage = new HelpFormatter();
		usage.printHelp("motras-algorithms", options);
	}

	private static void setOutput(Object output) {
		if (output instanceof State) {
			TransitionSystem transitionSystem = ((State) output).getContext();
			System.out.print(new XMTSWriter().write(transitionSystem));
		} else {
			System.out.println(output.toString());
		}
	}

}
