package edu.tum.motras.benchmarker;

import static com.google.common.base.Preconditions.checkArgument;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.NoSuchElementException;

import com.google.gson.Gson;

import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.utils.XMTSReader;
import edu.tum.xmts.utils.XMTSReader.MalformedInputException;
import edu.tum.xmts.utils.XMTSWriter;

public class Benchmark {

	public static class Pair {
		public final TransitionSystem leftSystem;
		public final TransitionSystem rightSystem;

		public Pair(TransitionSystem l, TransitionSystem r) {
			leftSystem = l;
			rightSystem = r;
		}

		public void writeTo(File file) throws IOException {
			try (BufferedWriter w = new BufferedWriter(new FileWriter(file))) {
				XMTSWriter xmtsWriter = new XMTSWriter();
				xmtsWriter.write(leftSystem, w);
				xmtsWriter.write(rightSystem, w);
			} catch (IllegalStateException e) {
				System.err.println("corrupted pair, dropping from test suite.");
				file.delete();
			}
		}

		public static Pair readFrom(File file) throws IOException, MalformedInputException {
			try (BufferedReader r = new BufferedReader(new FileReader(file))) {
				XMTSReader xmtsReader = new XMTSReader();
				List<TransitionSystem> systems = xmtsReader.readList(r);
				return new Pair(systems.get(0), systems.get(1));
			}
		}
	}

	public final SimplifiedXMTSGeneratorConfiguration parameters;
	public final int[] distanceVector;
	public final int limit;

	private transient List<Benchmark.Pair> True;
	private transient List<Benchmark.Pair> False;

	public Benchmark(SimplifiedXMTSGeneratorConfiguration p, int[] d, int l) {
		this.parameters = p;
		this.distanceVector = d;
		this.limit = l;

		this.True = new ArrayList<>(l);
		this.False = new ArrayList<>(l);
	}

	public void offerTrue(Benchmark.Pair pair) {
		if (True.size() < limit) {
			True.add(pair);
		}
	}

	public void offerFalse(Benchmark.Pair pair) {
		if (False.size() < limit) {
			False.add(pair);
		}
	}

	public boolean isComplete() {
		return (True.size() == limit && False.size() == limit);
	}

	public int totalSize() {
		return True.size() + False.size();
	}

	public boolean totalReachedLimit() {
		return (limit <= True.size() + False.size());
	}

	public List<Benchmark.Pair> getTrue() {
		return Collections.unmodifiableList(True);
	}

	public List<Benchmark.Pair> getFalse() {
		return Collections.unmodifiableList(False);
	}

	public void writeTo(File folder) throws IOException {
		checkArgument(folder.isDirectory());

		File config = new File(folder, "benchmarkSet.json");
		
		try (BufferedWriter w = new BufferedWriter(new FileWriter(config))) {
			Gson gson = new Gson();
			gson.toJson(this, w);
		}

		File trueFolder = new File(folder, "True");
		File falseFolder = new File(folder, "False");

		if (!trueFolder.mkdir() || !falseFolder.mkdir()) {
			throw new IOException("Could not create folders.");
		}

		for (int i = 0; i < True.size(); i++) {
			True.get(i).writeTo(new File(trueFolder, i + ".xmts"));
		}

		for (int i = 0; i < False.size(); i++) {
			False.get(i).writeTo(new File(falseFolder, i + ".xmts"));
		}
	}

	public static Benchmark readFrom(File folder) throws FileNotFoundException, IOException {
		checkArgument(folder.isDirectory());

		Benchmark benchmark;

		File config = new File(folder, "benchmarkSet.json");
		try (BufferedReader r = new BufferedReader(new FileReader(config))) {
			Gson gson = new Gson();
			benchmark = gson.fromJson(r, Benchmark.class);
		}

		File trueFolder = new File(folder, "True");
		File falseFolder = new File(folder, "False");

		benchmark.True = new ArrayList<>(benchmark.limit);
		benchmark.False = new ArrayList<>(benchmark.limit);

		for (File pair : trueFolder.listFiles((dir, name) -> name.endsWith(".xmts"))) {
			try {
				benchmark.True.add(Pair.readFrom(pair));
			} catch (NoSuchElementException | MalformedInputException | NullPointerException e) {
				System.err.println("corrupted pair (" + e.getMessage() + "), dropping from test suite.");
				pair.delete();
			}
		}

		for (File pair : falseFolder.listFiles((dir, name) -> name.endsWith(".xmts"))) {
			try {
				benchmark.False.add(Pair.readFrom(pair));
			} catch (NoSuchElementException | MalformedInputException e) {
				System.err.println("corrupted pair (" + e.getMessage() + "), dropping from test suite.");
				pair.delete();
			}
		}

		return benchmark;
	}
}