package edu.tum.xmts;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;

import edu.tum.xmts.Obligation.All;
import edu.tum.xmts.Obligation.DisjunctiveMust;
import edu.tum.xmts.Obligation.Generic;
import edu.tum.xmts.Obligation.Must;
import edu.tum.xmts.Obligation.Tran;
import edu.tum.xmts.formula.CheckContextVisitor;
import edu.tum.xmts.formula.Formula;
import edu.tum.xmts.formula.Formula.Atom;
import edu.tum.xmts.formula.Formula.Composite;
import edu.tum.xmts.formula.Formula.Conjunction;
import edu.tum.xmts.formula.Formula.Disjunction;

@NonNullByDefault
public class ObligationFactory {

	public static Obligation create(State state) {
		return createAll(state);
	}

	public static Obligation create(State state, Formula formula) {
		Obligation obligation;

		formula.accept(new CheckContextVisitor(state));

		/* Assume it's a MTS obligation */
		obligation = createMust(state, formula);

		if (obligation != null) {
			return obligation;
		}

		/* Assume it's a DMTS obligation */
		obligation = createDisjunctiveMust(state, formula);

		if (obligation != null) {
			return obligation;
		}

		/* Formula doesn't have the right form, falling back to Generic */
		return createGeneric(state, formula);
	}

	public static Obligation create(State state, Set<Set<Transition>> tran) {
		return new Tran(state, tran);
	}

	public static All createAll(State state) {
		return new All(state);
	}

	public static @Nullable DisjunctiveMust createDisjunctiveMust(State state, Formula formula) {
		/* |DisjunctiveMust| = 1 */
		if (formula instanceof Disjunction) {
			Collection<Transition> transitions = unpack((Disjunction) formula);

			if (transitions != null) {
				return new DisjunctiveMust(state, Collections.singleton(transitions));
			}

			return null;
		}

		/* |DisjunctiveMust| > 1 */
		if (!(formula instanceof Conjunction)) {
			return null;
		}

		List<Collection<Transition>> disjunctiveMustTransitions = new LinkedList<>();

		for (Formula childFormula : ((Conjunction) formula).getChildren()) {
			Transition transition = unpack(childFormula);

			if (transition != null) {
				disjunctiveMustTransitions.add(Collections.singleton(transition));
				continue;
			}

			if (!(childFormula instanceof Disjunction)) {
				return null;
			}

			Set<Transition> transitions = unpack((Disjunction) childFormula);

			if (transitions == null) {
				return null;
			}

			disjunctiveMustTransitions.add(transitions);
		}

		return new DisjunctiveMust(state, disjunctiveMustTransitions);
	}

	public static Generic createGeneric(State state, Formula formula) {
		return new Generic(state, formula);
	}

	/* The collections library is missing the necessary null-annotations */
	@SuppressWarnings("null")
	public static @Nullable Must createMust(State state, Formula formula) {
		/* |Must| = 0 */
		if (formula == Formula.TRUE) {
			return new Must(state, Collections.emptyList());
		}

		/* |Must| = 1 */
		Transition transition = unpack(formula);

		if (transition != null) {
			return new Must(state, Collections.singleton(transition));
		}

		/* |Must| > 1 */
		if (!(formula instanceof Conjunction)) {
			return null;
		}

		Set<Transition> mustTransitions = unpack((Conjunction) formula);

		if (mustTransitions != null) {
			return new Must(state, mustTransitions);
		}

		return null;
	}

	protected static @Nullable Set<Transition> unpack(Composite composite) {
		Set<Transition> transitions = new HashSet<>(composite.getChildren().size());
		Transition transition;

		for (Formula formula : composite.getChildren()) {
			transition = unpack(formula);

			if (transition == null) {
				return null;
			}

			transitions.add(transition);
		}

		return transitions;
	}

	protected static @Nullable Transition unpack(Formula formula) {
		if (formula instanceof Atom<?>) {
			Object payload = ((Atom<?>) formula).getPayload();

			if (payload instanceof Transition) {
				return (Transition) payload;
			}
		}

		return null;
	}
}
