package edu.tum.xmts.formula;

import static com.google.common.base.Preconditions.checkNotNull;

import java.lang.ref.WeakReference;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.eclipse.jdt.annotation.NonNull;
import org.eclipse.jdt.annotation.Nullable;

import com.google.common.collect.ImmutableList;

public abstract class Formula {

	public static final Formula FALSE = new Atom<>(false);
	public static final Formula TRUE = new Atom<>(true);

	public static class Atom<T> extends Formula {
		private final WeakReference<T> payload;

		public Atom(T payload) {
			checkNotNull(payload);
			this.payload = new WeakReference<>(payload);
		}

		public T getPayload() {
			@Nullable
			T object = payload.get();

			if (object == null) {
				throw new RuntimeException("WeakReference is null, there is a memory management issue");
			}

			return object;
		}

		@SuppressWarnings("all")
		@Override
		public <R> R accept(Visitor<R> visitor) {
			return visitor.visit(this);
		}

		@SuppressWarnings("all")
		@Override
		public void accept(VoidVisitor visitor) {
			visitor.visit(this);
		}

		@SuppressWarnings("all")
		@Override
		public boolean accept(@NonNull BooleanVisitor visitor) {
			return visitor.visit(this);
		}
	}

	public static abstract class Composite extends Formula {
		protected final ImmutableList<Formula> children;

		protected Composite(Collection<Formula> children) {
			this.children = ImmutableList.copyOf(children);
		}

		protected Composite(Formula... children) {
			this.children = ImmutableList.copyOf(children);
		}

		public List<Formula> getChildren() {
			return children;
		}
	}

	public static class Conjunction extends Composite {
		public Conjunction(Collection<Formula> children) {
			super(children);
		}

		public Conjunction(Formula... children) {
			super(children);
		}

		@Override
		public <R> R accept(Visitor<R> visitor) {
			return visitor.visit(this);
		}

		@Override
		public void accept(VoidVisitor visitor) {
			visitor.visit(this);
		}

		@Override
		public boolean accept(BooleanVisitor visitor) {
			return visitor.visit(this);
		}
	}

	public static class Disjunction extends Composite {
		public Disjunction(Collection<Formula> children) {
			super(children);
		}

		public Disjunction(Formula... children) {
			super(children);
		}

		@Override
		public <R> R accept(Visitor<R> visitor) {
			return visitor.visit(this);
		}

		@Override
		public void accept(VoidVisitor visitor) {
			visitor.visit(this);
		}

		@Override
		public boolean accept(BooleanVisitor visitor) {
			return visitor.visit(this);
		}
	}

	public static class Equivalence extends Composite {
		public Equivalence(Formula leftChild, Formula rightChild) {
			super(leftChild, rightChild);
		}

		@Override
		public <R> R accept(Visitor<R> visitor) {
			return visitor.visit(this);
		}

		@Override
		public void accept(VoidVisitor visitor) {
			visitor.visit(this);
		}

		@Override
		public boolean accept(BooleanVisitor visitor) {
			return visitor.visit(this);
		}
	}

	public static class Implication extends Composite {
		public Implication(Formula leftChild, Formula rightChild) {
			super(leftChild, rightChild);
		}

		@Override
		public <R> R accept(Visitor<R> visitor) {
			return visitor.visit(this);
		}

		@Override
		public void accept(VoidVisitor visitor) {
			visitor.visit(this);
		}

		@Override
		public boolean accept(BooleanVisitor visitor) {
			return visitor.visit(this);
		}
	}

	public static class Negation extends Composite {
		public Negation(Formula child) {
			super(child);
		}

		@Override
		public <R> R accept(Visitor<R> visitor) {
			return visitor.visit(this);
		}

		@Override
		public void accept(VoidVisitor visitor) {
			visitor.visit(this);
		}

		@Override
		public boolean accept(BooleanVisitor visitor) {
			return visitor.visit(this);
		}
	}

	public static Formula and(Collection<Formula> children) {
		if (children.isEmpty()) {
			return Formula.TRUE;
		}

		if (children.size() == 1) {
			return children.iterator().next();
		}

		return new Conjunction(children);
	}

	public static Formula and(Formula... children) {
		@NonNull
		List<Formula> asList = Arrays.asList(children);
		return (and(asList));
	}

	public static Formula or(Collection<Formula> children) {
		if (children.isEmpty()) {
			return Formula.FALSE;
		}

		if (children.size() == 1) {
			return children.iterator().next();
		}

		return new Disjunction(children);
	}

	public static Formula or(Formula... children) {
		return (or(Arrays.asList(children)));
	}

	abstract public <R> R accept(Visitor<R> visitor);

	abstract public boolean accept(BooleanVisitor visitor);

	abstract public void accept(VoidVisitor visitor);

	public Conjunction and(Formula formula) {
		return new Conjunction(this, formula);
	}

	public Disjunction or(Formula formula) {
		return new Disjunction(this, formula);
	}

	public Equivalence iff(Formula formula) {
		return new Equivalence(this, formula);
	}

	public Formula implies(Formula formula) {
		return new Implication(this, formula);
	}

	public Negation not() {
		return new Negation(this);
	}
}
