package edu.tum.xmts.utils;

import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Random;
import java.util.Set;

import com.google.common.collect.Sets;

import edu.tum.xmts.Action;
import edu.tum.xmts.Parameter;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.formula.Formula;
import edu.tum.xmts.formula.SimplifyVisitor;

public class XMTSGenerator {

	public static class XMTSGeneratorConfiguration {
		/* Suppress weird warning, that cannot be fixed */
		@SuppressWarnings("all")
		public Type type = Type.MTS;
		public int parameters;
		public int actions;
		public int clusterInterfaceNodes;
		public int clusterNodeDegree;
		public int clusterSize;
		public int systemNodeDegree;
		public int systemSize;
		public boolean deterministic;
	}

	private final Random structureRandomness;
	private final Random obligationRandomness;

	public XMTSGenerator() {
		structureRandomness = new Random();
		obligationRandomness = new Random();
	}

	private void generateStructure(TransitionSystem transitionSystem, XMTSGeneratorConfiguration configuration) {
		/* Create Parameters */
		for (int i = 0; i < configuration.parameters; i++) {
			transitionSystem.parameters.create();
		}

		/* Create Actions */
		for (int i = 0; i < configuration.actions; i++) {
			transitionSystem.actions.create();
		}

		/* Create States and Transitions */
		Set<State> interfaceNodes = new HashSet<>(configuration.systemSize * configuration.clusterInterfaceNodes);
		State root;

		for (int i = 0; i < configuration.systemSize; i++) {
			interfaceNodes.addAll(generateCluster(transitionSystem, configuration));
		}

		root = generateStructure(interfaceNodes, configuration.systemNodeDegree + configuration.clusterNodeDegree,
				configuration.deterministic);
		transitionSystem.setInitialState(root);
	}

	private void generateObligation(TransitionSystem transitionSystem, XMTSGeneratorConfiguration configuration) {
		Collection<Transition> mustTransitions;
		Collection<Collection<Transition>> disjunctiveMust;

		for (State state : transitionSystem.states.getElements()) {
			int transitions = obligationRandomness.nextInt(state.getTransitions().size() + 1);

			switch (configuration.type) {
			case MTS:
				mustTransitions = RandomUtil.randomSubList(obligationRandomness, state.getTransitions(), transitions);
				state.setMustObligation(mustTransitions);
				break;

			case DMTS:
				disjunctiveMust = new LinkedList<>();

				for (int i = 0; i < 5; i++) {
					mustTransitions = RandomUtil.randomSubList(obligationRandomness, state.getTransitions(),
							transitions / 2);

					if (mustTransitions.isEmpty()) {
						continue;
					}

					disjunctiveMust.add(mustTransitions);
				}

				state.setDisjunctiveMustObligation(disjunctiveMust);
				break;

			case BMTS:
			case PMTS:
				state.setObligation(generateGenericObligation(transitionSystem, state));
				break;

			default:
				state.setAllObligation();
			}
		}
	}

	public TransitionSystem generate(XMTSGeneratorConfiguration configuration, long structureSeed, long obligationSeed) {
		structureRandomness.setSeed(structureSeed);
		obligationRandomness.setSeed(obligationSeed);

		TransitionSystem transitionSystem = new TransitionSystem("random_" + structureSeed + "_" + obligationSeed);
		transitionSystem.setType(configuration.type);

		generateStructure(transitionSystem, configuration);
		generateObligation(transitionSystem, configuration);

		return transitionSystem;
	}

	public TransitionSystem generate(XMTSGeneratorConfiguration configuration, long seed) {
		return generate(configuration, seed, seed + 1);
	}

	private Formula generateGenericObligation(TransitionSystem system, State state) {
		Deque<Formula> formulae = new LinkedList<>();

		formulae.add(Formula.TRUE);

		int transitions = obligationRandomness.nextInt(state.getTransitions().size() + 1);

		for (Transition transition : RandomUtil
				.randomSubList(obligationRandomness, state.getTransitions(), transitions)) {
			formulae.add(new Formula.Atom<>(transition));
		}

		int parameters = obligationRandomness.nextInt(3);

		for (Parameter parameter : RandomUtil.randomSubList(obligationRandomness, system.parameters.getElements(),
				parameters)) {
			formulae.add(new Formula.Atom<>(parameter));
		}

		while (formulae.size() > 1) {
			int r = obligationRandomness.nextInt(5);
			Set<Formula> children = new HashSet<>();

			switch (r) {
			case 0:
				r = obligationRandomness.nextInt(formulae.size());
				for (; 0 < r; r--) {
					children.add(formulae.removeFirst());
				}
				formulae.addLast(Formula.and(children));
				break;

			case 1:
				r = obligationRandomness.nextInt(formulae.size());
				for (; 0 < r; r--) {
					children.add(formulae.removeFirst());
				}
				formulae.addLast(Formula.or(children));
				break;

			case 2:
				formulae.addLast(formulae.removeFirst().not());
				break;

			case 3:
				formulae.addLast(formulae.removeFirst().implies(formulae.removeFirst()));
				break;

			case 4:
				formulae.addLast(formulae.removeFirst().iff(formulae.removeFirst()));
				break;
			}
		}

		SimplifyVisitor visitor = new SimplifyVisitor();
		Formula formula = formulae.removeFirst();
		/* Suppress weird warning, that cannot be fixed */
		@SuppressWarnings("all")
		Formula formula2 = formula.accept(visitor);
		return formula2;
	}

	private Set<State> generateCluster(TransitionSystem system, XMTSGeneratorConfiguration configuration) {
		State clusterRoot;
		Set<State> cluster = new HashSet<>(configuration.clusterSize);
		Set<State> interfaceNodes = new HashSet<>(configuration.clusterInterfaceNodes);

		for (int i = 0; i < configuration.clusterSize; i++) {
			cluster.add(system.states.create());
		}

		// Build transitions and ensure that clusterRoot is contained in
		// interfaceNodes
		clusterRoot = generateStructure(cluster, configuration.clusterNodeDegree, configuration.deterministic);
		cluster.remove(clusterRoot);
		interfaceNodes.add(clusterRoot);
		interfaceNodes.addAll(RandomUtil.randomSubList(structureRandomness, cluster,
				configuration.clusterInterfaceNodes - 1));
		return interfaceNodes;
	}

	private State generateStructure(Set<State> states, int degree, boolean deterministic) {
		Deque<State> waiting = new LinkedList<>();
		Deque<State> todo = new LinkedList<>(states);
		State root = todo.removeFirst();
		waiting.addLast(root);

		while (!waiting.isEmpty()) {
			State current = waiting.removeFirst();

			// Build tree-like structure
			for (int i = 0; i < degree - 1 && !todo.isEmpty(); i++) {
				State target = todo.removeFirst();

				Set<State> singleton = Collections.singleton(target);

				generateTransition(current, singleton, deterministic);
				waiting.addLast(target);
			}
		}

		// Add some fuzziness
		for (State state : states) {
			while (state.getTransitions().size() < degree) {
				generateTransition(state, states, deterministic);
			}
		}

		return root;
	}

	private void generateTransition(State state, Set<State> cluster, boolean deterministic) {
		Action action;
		State target;

		for (int i = 0; i < 5; i++) {
			if (cluster.size() == 1) {
				target = cluster.iterator().next();
			} else {
				target = RandomUtil.randomElement(structureRandomness, cluster);
			}

			if (!deterministic) {
				action = RandomUtil.randomElement(structureRandomness, state.getContext().actions.getElements());
			} else {
				Set<Action> difference = Sets.difference(new HashSet<>(state.getContext().actions.getElements()),
						new HashSet<>(state.getActions()));

				switch (difference.size()) {
				case 0:
					return;

				case 1:
					action = difference.iterator().next();
					break;

				default:
					action = RandomUtil.randomElement(structureRandomness, difference);
				}
			}

			if (state.getTransition(action, target) == null) {
				state.createTransition(action, target);
				return;
			}
		}
	}

}