package edu.tum.xmts.utils;

import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;

import org.eclipse.jdt.annotation.Nullable;

import edu.tum.xmts.Action;
import edu.tum.xmts.Obligation;
import edu.tum.xmts.Parameter;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.formula.Formula;
import edu.tum.xmts.formula.Formula.Atom;
import edu.tum.xmts.formula.Formula.Composite;
import edu.tum.xmts.formula.Formula.Conjunction;
import edu.tum.xmts.formula.Formula.Disjunction;
import edu.tum.xmts.formula.Formula.Equivalence;
import edu.tum.xmts.formula.Formula.Implication;
import edu.tum.xmts.formula.Formula.Negation;
import edu.tum.xmts.formula.Visitor;

public class XMTSWriter implements TransitionSystemWriter {

	// Obvious performance impact, but just keep it simple
	protected class FormulaConverter implements Visitor<String> {

		@Override
		public String visit(Atom<?> atom) {
			Object payload = atom.getPayload();

			if (payload instanceof Parameter) {
				return ((Parameter) payload).getId();
			}

			if (payload instanceof Transition) {
				return ((Transition) payload).getAction() + "." + ((Transition) payload).getState();
			}

			if (atom == Formula.TRUE) {
				return "true";
			}

			if (atom == Formula.FALSE) {
				return "false";
			}

			throw new IllegalStateException();
		}

		@SuppressWarnings("null")
		private String visitComposite(Composite composite, String symbol) {
			StringWriter w = new StringWriter();
			int size = composite.getChildren().size();

			w.write("(");

			for (int i = 0; i < size - 1; i++) {
				w.write(composite.getChildren().get(i).accept(this));
				w.write(" " + symbol + " ");
			}

			w.write(composite.getChildren().get(size - 1).accept(this));
			w.write(")");

			return w.toString();
		}

		@Override
		public String visit(Conjunction conjunction) {
			return visitComposite(conjunction, "&");
		}

		@Override
		public String visit(Disjunction disjunction) {
			return visitComposite(disjunction, "|");
		}

		@Override
		public String visit(Equivalence equivalence) {
			return visitComposite(equivalence, "<->");
		}

		@Override
		public String visit(Implication implication) {
			return visitComposite(implication, "->");
		}

		@Override
		public String visit(Negation negation) {
			return "(!" + negation.getChildren().get(0).accept(this) + ")";
		}

	}

	public String indentation = "    ";

	@SuppressWarnings("null")
	@Override
	public String write(TransitionSystem ts) {
		StringWriter stringWriter = new StringWriter();

		try {
			write(ts, stringWriter);
		} catch (IOException e) {
			e.printStackTrace();
		}

		return stringWriter.toString();
	}

	@Override
	public void write(TransitionSystem ts, Writer w) throws IOException {
		writeLine(w, 0, ts.getType().toString(), ts.getId(), true);
		writeLine(w, 1, "label", ts.getLabel(), false);
		writeLine(w, 1, "initial_state", ts.getInitialState(), false);

		for (Parameter parameter : ts.parameters.getElements()) {
			write(parameter, w);
		}

		w.write("\n");

		for (Action action : ts.actions.getElements()) {
			write(action, w);
		}

		w.write("\n");

		for (State state : ts.states.getElements()) {
			write(state, w);
		}

		writeLine(w, 0, true);
		w.flush();
	}

	@SuppressWarnings("null")
	public String writeFormula(Formula formula) {
		StringWriter stringWriter = new StringWriter();

		try {
			writeFormula(formula, stringWriter);
		} catch (IOException e) {
			e.printStackTrace();
		}

		return stringWriter.toString();
	}

	public void writeFormula(Formula formula, Writer w) throws IOException {
		w.write((formula).accept(new FormulaConverter()));
	}

	private void writeLine(Writer w, int indentLevel, String key, @Nullable Object value, boolean record)
			throws IOException {
		if (value == null) {
			return;
		}

		for (int i = indentLevel; i > 0; i--) {
			w.write(indentation);
		}

		w.write(key + " ");

		if (value instanceof String) {
			if (key.equals("label")) {
				w.write("\"");
			}
			w.write((String) value);
			if (key.equals("label")) {
				w.write("\"");
			}
		} else if (value instanceof Formula) {
			writeFormula((Formula) value, w);
		} else if (value instanceof Obligation.All) {
			w.write("all");
		} else if (value instanceof Obligation) {
			w.write(((Obligation) value).getFormula().accept(new FormulaConverter()));
		} else {
			w.write(value.toString());
		}

		w.write(record ? " [\n" : "\n");
	}

	private void writeLine(Writer w, int indentLevel, boolean record) throws IOException {
		for (int i = indentLevel; i > 0; i--) {
			w.write(indentation);
		}

		w.write(record ? "]\n" : "\n");
	}

	private void write(Parameter parameter, Writer w) throws IOException {
		boolean record = parameter.getLabel() != null || parameter.getInvestmentCost() != null;

		writeLine(w, 1, "parameter", parameter.getId(), record);
		writeLine(w, 2, "label", parameter.getLabel(), false);
		writeLine(w, 2, "investment_cost", parameter.getInvestmentCost(), false);
		writeLine(w, 1, record);
	}

	private void write(Action action, Writer w) throws IOException {
		boolean record = action.getLabel() != null || action.getRequirement() != null
				|| action.getRunningCost() != null;

		writeLine(w, 1, "action", action.getId(), record);
		writeLine(w, 2, "label", action.getLabel(), false);
		writeLine(w, 2, "requirement", action.getRequirement(), false);
		writeLine(w, 2, "running_cost", action.getRunningCost(), false);
		writeLine(w, 1, record);
	}

	private void write(State state, Writer w) throws IOException {
		boolean record = state.getLabel() != null || state.getPosition() != null || !state.getTransitions().isEmpty();

		writeLine(w, 1, "state", state.getId(), record);
		writeLine(w, 2, "label", state.getLabel(), false);
		writeLine(w, 2, "obligation", state.getObligation(), false);
		writeLine(w, 2, "position", state.getPosition(), false);

		for (Transition transition : state.getTransitions()) {
			write(transition, w);
		}

		writeLine(w, 1, record);
	}

	private void write(Transition transition, Writer w) throws IOException {
		boolean record = transition.getLabel() != null || transition.getDuration() != null;

		writeLine(w, 2, transition.getAction() + " -> " + transition.getState(), "", record);
		writeLine(w, 3, "label", transition.getLabel(), false);
		writeLine(w, 3, "duration", transition.getDuration(), false);
		writeLine(w, 2, record);
	}
}
